local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local LCo = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-Config")
local ACD = LibStub("AceConfigDialog-3.0")
local ACR = LibStub("AceConfigRegistry-3.0")

local CreateMacro, EditMacro, PickupMacro, GetMacroIndexByName = CreateMacro, EditMacro, PickupMacro, GetMacroIndexByName
local pairs, type = pairs, type

local optMGetter, optMSetter, optQGetter, optWQLSetter, optCCFSetter, optTagSetter, optHideBTr, specialItemMacro
do
	function optMGetter(info)
		return core:getModuleEnabled(info[#info])
	end
	function optMSetter(info, value)
		core:setModuleEnabled(info[#info], value)
	end
	function optQGetter(info)
		return core.db.profile.quests[info[#info]]
	end
	function optWQLSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		core:computeQuestLevel()
	end
	function optCCFSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		core:computeQuestsObjectivesStatus()
	end
	function optTagSetter(info, value)
		core.db.profile.quests[info[#info]] = value
		core:computeQuestTag()
	end
	function optHideBTr(info, value)
		core.db.profile.quests[info[#info]] = value
		core:hideShowDefaultBlizTracker()
	end
	function specialItemMacro(info, value)
		local index = GetMacroIndexByName("tomQuest2_SI")
		if index == 0 then
			index = CreateMacro("tomQuest2_SI", 1, "", 0)
			PickupMacro(index)
		else
			EditMacro(index, "tomQuest2_SI", 1, "")
		end
	end
end

local options, moduleOptions = nil, {}
local function getOptions()
	if not options then
		options = {
			type = "group",
			name = "tomQuest2",
			args = {
				modules = {
					type = "group",
					name = LCo["Modules"],
					desc = LCo["Modules"],
					get = optQGetter,
					args = {
						header = {
							type = "header",
							name = LCo["Modules"],
							order = 0,
						},
						hideDefaultBlizTracker = {
							name = LCo["Hide default blizzard quests and achievement tracker"],
							type= "toggle",
							set = optHideBTr,
							order = 100,
							width = "full",
						},
						weightedLevel = {
							type = "header",
							name = LCo["Weighted quest level computation"],
							order = 101,
						},
						eliteAddLevel = {
							name = LCo["Level to add for an elite quest"],
							type = "range",
							min = 0,
							max = 5,
							step = 1,
							set = optWQLSetter,
							order = 102,
						},
						groupAddLevel = {
							name = LCo["Level to add per missing player for a group quest"],
							type = "range",
							min = 0,
							max = 5,
							step = 1,
							set = optWQLSetter,
							order = 103,
						},
						dungeonAddLevel = {
							name = LCo["Level to add per missing player for a dungeon quest"],
							type = "range",
							min = 0,
							max = 5,
							step = 1,
							set = optWQLSetter,
							order = 104,
						},
						heroicAddLevel = {
							name = LCo["Level to add per missing player for an heroic quest"],
							type = "range",
							min = 0,
							max = 5,
							step = 1,
							set = optWQLSetter,
							order = 105,
						},
						questsTag = {
							type = "header",
							name = LCo["Custom quests tag"],
							order = 106,
						},
						elite = {
							name = LCo["Elite quest tag"],
							type = "input",
							set = optTagSetter,
							order = 107,
						},
						group = {
							name = LCo["Group quest tag"],
							type = "input",
							set = optTagSetter,
							order = 108,
						},
						pvp = {
							name = LCo["PVP quest tag"],
							type = "input",
							set = optTagSetter,
							order = 109,
						},
						raid = {
							name = LCo["Raid quest tag"],
							type = "input",
							set = optTagSetter,
							order = 110,
						},
						dungeon = {
							name = LCo["Dungeon quest tag"],
							type = "input",
							set = optTagSetter,
							order = 111,
						},
						heroic = {
							name = LCo["Heroic quest tag"],
							type = "input",
							set = optTagSetter,
							order = 112,
						},
						daily = {
							name = LCo["Daily quest tag"],
							type = "input",
							set = optTagSetter,
							order = 113,
						},
						macroHeader = {
							type = "header",
							name = LCo["Special quests item macro"],
							order = 114,
						},
						macro = {
							name = LCo["Macro Generator"],
							type = "execute",
							func = specialItemMacro,
							order = 115,
						},
						customColorFunctionHeader = {
							type = "header",
							name = LCo["Custom objective coloring function"],
							order = 116,
						},
						customColorFunctionDesc = {
							type = "description",
							name = LCo["Custom objective coloring allow you to define multiplicative and additive component of the coloring function. For each color component you'll have a function that is color = (multiplicative component) * percent + (additive component) where percent is the percent of completion of the objective."],
							order = 117,
						},
						zeroCustomColorFunctionHeader = {
							type = "header",
							name = LCo["Function for objective completion at 0%"],
							order = 118,
						},
						rj0 = {
							name = LCo["Red add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 119,
						},
						gj0 = {
							name = LCo["Green add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 120,
						},
						bj0 = {
							name = LCo["Blue add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 121
						},
						firstCustomColorFunctionHeader = {
							type = "header",
							name = LCo["Function for [0%, 50%[ objective completion"],
							order = 122,
						},
						ri1 = {
							name = LCo["Red mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 123,
						},
						gi1 = {
							name = LCo["Green mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 124,
						},
						bi1 = {
							name = LCo["Blue mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 125,
						},
						rj1 = {
							name = LCo["Red add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 126,
						},
						gj1 = {
							name = LCo["Green add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 127,
						},
						bj1 = {
							name = LCo["Blue add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 128,
						},
						secondCustomColorFunctionHeader = {
							type = "header",
							name = LCo["Function for ]50%, 100%[ objective completion"],
							order = 129,
						},
						ri2 = {
							name = LCo["Red mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 130,
						},
						gi2 = {
							name = LCo["Green mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 131,
						},
						bi2 = {
							name = LCo["Blue mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 132,
						},
						rj2 = {
							name = LCo["Red add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 133,
						},
						gj2 = {
							name = LCo["Green add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 134,
						},
						bj2 = {
							name = LCo["Blue add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 135,
						},
						fullCustomColorFunctionHeader = {
							type = "header",
							name = LCo["Function for objective completion at 100%"],
							order = 136,
						},
						ri3 = {
							name = LCo["Red mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 137,
						},
						gi3 = {
							name = LCo["Green mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 138,
						},
						bi3 = {
							name = LCo["Blue mult."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 139,
						},
						rj3 = {
							name = LCo["Red add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 140,
						},
						gj3 = {
							name = LCo["Green add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 141,
						},
						bj3 = {
							name = LCo["Blue add."],
							type = "range",
							min = -5,
							max = 5,
							step = 0.1,
							set = optCCFSetter,
							order = 142,
						},
					},
				},
			},
		}
		local moduleIndex = 0
		for k, v in core:IterateModules() do
			moduleIndex = moduleIndex + 1
			options.args.modules.args[k] = {
				type = "toggle",
				name = v.modName,
				get = optMGetter,
				set = optMSetter,
				width = "double",
				desc = v.modName,
				order = moduleIndex,
			}
		end
		for k,v in pairs(moduleOptions) do
			options.args[k] = (type(v) == "function") and v() or v
		end
	end
	return options
end

function core:SetupOptions()
	InterfaceOptionsFrame:SetFrameStrata("DIALOG") 
	core.optionsFrames = {}
	ACR:RegisterOptionsTable("tomQuest2", getOptions)
	self.optionsFrames.tomQuest2 = ACD:AddToBlizOptions("tomQuest2", "tomQuest2", nil, "modules")
	self:registerModuleOptions("Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db), LCo["Profiles"])
end

function core:registerModuleOptions(name, optionTbl, displayName)
	moduleOptions[name] = optionTbl
	self.optionsFrames[name] = ACD:AddToBlizOptions("tomQuest2", displayName, "tomQuest2", name)
end