-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("informations", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LI = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsInformations")
local LSM = LibStub("LibSharedMedia-3.0")

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local format, gsub, type = format, gsub, type
local abs, floor, mod, tonumber = math.abs, math.floor, mod, tonumber
local unpack, tinsert, tremove, pairs = unpack, tinsert, tremove, pairs
local CreateFrame, CreateFont = CreateFrame, CreateFont
local GetQuestLogSelection, SelectQuestLogEntry = GetQuestLogSelection, SelectQuestLogEntry
local QuestLogPushQuest = QuestLogPushQuest
local SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems = SetAbandonQuest, GetAbandonQuestName, GetAbandonQuestItems
local IsControlKeyDown, IsShiftKeyDown, IsAltKeyDown = IsControlKeyDown, IsShiftKeyDown, IsAltKeyDown
local GetItemInfo, DressUpItemLink, SetItemRef = GetItemInfo, DressUpItemLink, SetItemRef

-----------------
--- variables ---
-----------------
module.modName = LI["Quests Informations"]
local _
local prdb, options
local data, stack = {}, {}
local COLOR_COPPER = "eda55f"
local COLOR_SILVER = "c7c7cf"
local COLOR_GOLD = "ffd700"
local attachSelect = {[1] = LI["Right"], [2] = LI["Left"], [3] = LI["Dettached"]}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showQuestLog = true,
		showLightHeaded = true,
		backDropColor = {0.09, 0.09, 0.19, 0.5},
		borderColor = {0.5, 0.5, 0.5, 1},
		attach = 1,
		showAllCommentsOnOnePage = false,
		qlFont = LSM.DefaultMedia.font,
		qlFontSizeH1 = 14,
		qlFontSizeH2 = 12,
		qlFontSize = 10,
		lhFont = LSM.DefaultMedia.font,
		lhFontSize = 10,
		tomQuest2ParentFrame = {
			width = 640,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2QlFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		tomQuest2LhFrame = {
			width = 320,
			height = 400,
			point = "CENTER",
			x = 0,
			y = 0,
		},
		color = "ffffd100",
	},
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, frameSetter, hideLH, hideColor, hideAttach, hideQL, hideLH2
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
		local uid
		if tomQuest2QlFrame then
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
		local uid
		if tomQuest2ParentFrame then
			tomQuest2ParentFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2ParentFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		end
		if tomQuest2QlFrame then
			tomQuest2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2QlFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			module:lockUnlockQlFrame()
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			tomQuest2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tomQuest2LhFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			module:lockUnlockLhFrame()
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
	function frameSetter(info, value)
		prdb[info[#info]] = value
		local uid
		if tomQuest2QlFrame then
			module:lockUnlockQlFrame()
			uid = tomQuest2QlFrame.currentUid
			tomQuest2QlFrame.currentUid = nil
		end
		if tomQuest2LhFrame then
			module:lockUnlockLhFrame()
			uid = tomQuest2LhFrame.currentUid or uid
			tomQuest2LhFrame.currentUid = nil
		end
		if uid then
			module:displayQuestInformations(nil, uid)
		end
	end
	function hideLH()
		if LightHeaded then
			return false
		else
			return true
		end
	end
	function hideColor()
		if prdb.showQuestLog or prdb.showLightHeaded then
			return false
		else
			return true
		end
	end
	function hideAttach()
		if LightHeaded and prdb.showQuestLog and prdb.showLightHeaded then
			return false
		else
			return true
		end
	end
	function hideQL()
		if prdb.showQuestLog then
			return false
		else
			return true
		end
	end
	function hideLH2()
		if LightHeaded and prdb.showLightHeaded then
			return false
		else
			return true
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				showQuestLog = {
					name = LI["Show tomQuest2 quest log"],
					type = "toggle",
					set = frameSetter,
					width = "full",
					order = 4,
				},
				showLightHeaded = {
					name = LI["Show LightHeaded into tomQuest2"],
					hidden = hideLH,
					type = "toggle",
					set = frameSetter,
					width = "full",
					order = 5,
				},
				backDropColor = {
					name = LI["Background color"],
					hidden = hideColor,
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 6,
				},
				borderColor = {
					name = LI["Border color"],
					hidden = hideColor,
					type = "color",
					hasAlpha = true,
					get = colorGetter,
					set = colorSetter,
					width = "full",
					order = 7,
				},
				attach = {
					name = LI["Attach LightHeaded to QuestLog"],
					hidden = hideAttach,
					type = "select",
					values = attachSelect,
					set = frameSetter,
					width = "full",
					order = 8,
				},
				showAllCommentsOnOnePage = {
					name = LI["Show all LightHeaded comments on the same page"],
					hidden = hideQL,
					type = "toggle",
					width = "full",
					order = 9,
				},
				questLogHeader = {
					name = LI["Quests log settings"],
					hidden = hideQL,
					type = "header",
					order = 10,
				},
				qlFont = {
					name = LI["Font"],
					hidden = hideQL,
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 11,
				},
				qlFontSizeH1 = {
					name = LI["Quest title font size"],
					hidden = hideQL,
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 12,
				},
				qlFontSizeH2 = {
					name = LI["Headers font size"],
					hidden = hideQL,
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 13,
				},
				qlFontSize = {
					name = LI["text font size"],
					hidden = hideQL,
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 13,
				},
				lightHeadedHeader = {
					name = LI["LightHeaded settings"],
					hidden = hideLH2,
					type = "header",
					order = 14,
				},
				lhFont = {
					name = LI["Font"],
					hidden = hideLH2,
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					width = "full",
					order = 15,
				},
				lhFontSize = {
					name = LI["Font size"],
					hidden = hideLH2,
					type="range",
					min = 6,
					max = 20,
					step = 1,
					width = "full",
					order = 16,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("informations", defaults, "Default")
	prdb = self.db.profile
	self:SetEnabledState(core:getModuleEnabled("informations"))
	core:registerModuleOptions("informations", getOptions, module.modName)
end

function module:OnEnable()
	core.RegisterCallback(self, "displayQuestInformations")
	core.RegisterCallback(self, "playerQuestLost", "questLost")
	core.RegisterCallback(self, "partyQuestLost", "questLost")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLost(event, first, commHandler, uid, abandoning, title, sender)
	if core.quests[uid] then return end
	local tq2QlFrame = tomQuest2QlFrame
	local tq2LhFrame = tomQuest2LhFrame
	local tq2ParentFrame = tomQuest2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if tq2ParentFrame and tq2ParentFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2ParentFrame:Hide()
			tq2QlFrame.currentUid = nil
			tq2LhFrame.currentUid = nil
			return
		end
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2QlFrame:Hide()
			tq2QlFrame.currentUid = nil
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() and tq2LhFrame.currentUid and tq2LhFrame.currentUid == uid then
			tq2LhFrame:Hide()
			tq2LhFrame.currentUid = nil
			return
		end
	end
end

-----------------------
--- Frame Functions ---
-----------------------
function module:createLhGUI()
	if tomQuest2LhFrame then return end
	local tq2LhFrame = CreateFrame("Frame", "tomQuest2LhFrame", UIParent)
	tinsert(UISpecialFrames, "tomQuest2LhFrame")
	tq2LhFrame:SetWidth(640)
	tq2LhFrame:SetHeight(512)
	tq2LhFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])

	local close = CreateFrame("Button", nil, tq2LhFrame, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT")
	close:SetScript("OnClick", function()
		tq2LhFrame.currentUid = nil
		tq2LhFrame:Hide()
	end)

	local resize = CreateFrame("Button", nil, tq2LhFrame)
	resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHeight(12)
	resize:SetWidth(12)
	resize:SetPoint("BOTTOMRIGHT", -3, 3)
	self:setFrameTooltip(resize, LI["Click to resize tomQuest2"])

	local titlereg = CreateFrame("Button", nil, tq2LhFrame)
	titlereg:SetPoint("TOPLEFT", 5, -5)
	titlereg:SetPoint("TOPRIGHT", 0, 0)
	titlereg:SetHeight(20)
	titlereg:SetScript("OnMouseDown", function(frame)
		tq2LhFrame:StartMoving()
	end)
	titlereg:SetScript("OnMouseUp", function(frame)
		tq2LhFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2LhFrame")
	end)

	tq2LhFrame:EnableMouse()
	tq2LhFrame:SetMovable(1)
	tq2LhFrame:SetResizable(1)
	tq2LhFrame:SetMinResize(300, 300)
	tq2LhFrame:SetWidth(400)
	tq2LhFrame:SetFrameLevel(0)
	tq2LhFrame:SetWidth(325)
	tq2LhFrame:SetHeight(450)

	tq2LhFrame.close = close
	tq2LhFrame.resize = resize
	tq2LhFrame.titlereg = titlereg

	tq2LhFrame.icon = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.icon:SetHeight(64)
	tq2LhFrame.icon:SetWidth(64)
	tq2LhFrame.icon:SetNormalTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon")
	tq2LhFrame.icon:SetHighlightTexture("Interface\\AddOns\\LightHeaded\\images\\wh_icon_hover")
	tq2LhFrame.icon:GetHighlightTexture():SetBlendMode("BLEND")
	tq2LhFrame.icon:SetPoint("BOTTOMRIGHT", -10, -10)

	StaticPopupDialogs["LIGHTHEADED_ABOUT_DIALOG"] = {
		text = "LightHeaded (C) 2007-2008 by Jim Whitehead\n\nThis add-on only provides you with the comments and some basic information on quests. For the full information, be sure to visit http://www.wowhead.com",
		button1 = TEXT(OKAY),
		OnAccept = function()end,
		timeout = 0,
		hideOnEscape = 1
	}

	tq2LhFrame.icon:SetScript("OnClick", function()
		StaticPopup_Show("LIGHTHEADED_ABOUT_DIALOG")
	end)

	tq2LhFrame.page = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.page:SetFontObject(GameFontNormalSmall)
	tq2LhFrame.page:SetPoint("BOTTOM", -32, 15)

	tq2LhFrame.prev = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.prev:SetWidth(21)
	tq2LhFrame.prev:SetHeight(21)
	tq2LhFrame.prev:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up")
	tq2LhFrame.prev:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down")
	tq2LhFrame.prev:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled")
	tq2LhFrame.prev:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.prev:SetPoint("RIGHT", tq2LhFrame.page, "LEFT", -25, 0)

	self:setFrameTooltip(tq2LhFrame.prev, LI["Previous page"])
	tq2LhFrame.prev:SetScript("OnClick", function(frame)
		if tq2LhFrame.currentPage then
			self:updateQuestsLightHeadedInformations(tq2LhFrame.currentUid, tq2LhFrame.currentPage - 1)
		end
	end)

	tq2LhFrame.next = CreateFrame("Button", nil, tq2LhFrame)
	tq2LhFrame.next:SetWidth(21)
	tq2LhFrame.next:SetHeight(21)
	tq2LhFrame.next:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up")
	tq2LhFrame.next:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down")
	tq2LhFrame.next:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled")
	tq2LhFrame.next:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
	tq2LhFrame.next:SetPoint("LEFT", tq2LhFrame.page, "RIGHT", 25, 0)

	self:setFrameTooltip(tq2LhFrame.next, LI["Next page"])
	tq2LhFrame.next:SetScript("OnClick", function(frame)
		if tq2LhFrame.currentPage then
			self:updateQuestsLightHeadedInformations(tq2LhFrame.currentUid, tq2LhFrame.currentPage + 1)
		end
	end)

	tq2LhFrame.title = tq2LhFrame:CreateFontString(nil, "ARTWORK")
	tq2LhFrame.title:SetPoint("TOP", 0, -4)

	tq2LhFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2LhScrollFrame", tq2LhFrame, "UIPanelScrollFrameTemplate")
	tq2LhFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2LhFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2LhFrame.tooltip = CreateFrame("GameTooltip", "tomQuest2LhTooltip", UIParent, "GameTooltipTemplate")

	tq2LhFrame.text = CreateFrame("SimpleHTML", "tomQuest2LhSimpleHTML", tq2LhFrame.scroll)
	tq2LhFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2LhFrame.text:SetHeight(300)
	tq2LhFrame.text:SetFontObject(GameFontHighlight)

	local function tq2LhOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2LhFrame.text:mySetText(msg)
		self.text = msg
		tq2LhFrame.updateText = msg
		tq2LhFrame:SetScript("OnUpdate", tq2LhOnUpdateFix)
	end

	function tq2LhFrame.text:updateSize()
		local lhw = tomQuest2LhFrame:GetWidth()
		local width = lhw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2LhFrame.text:SetScript("OnHyperlinkClick", function(...) self:onHyperlinkClick(...) end)
	tq2LhFrame.text:SetScript("OnHyperlinkEnter", function(...) self:onHyperlinkEnter(...) end)
	tq2LhFrame.text:SetScript("OnHyperlinkLeave", function() tq2LhFrame.tooltip:Hide() end)
	tq2LhFrame.scroll:SetScrollChild(tq2LhFrame.text)
	resize:SetScript("OnMouseDown", function(frame)
		tomQuest2LhFrame:StartSizing()
	end)
	resize:SetScript("OnMouseUp", function(frame)
		tomQuest2LhFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2LhFrame")
		tq2LhFrame.text:updateSize()
	end)
	tq2LhFrame:SetAlpha(1)
	tq2LhFrame:Show()
	self:lockUnlockLhFrame()
end

function module:createQlGUI()
	if tomQuest2QlFrame then return end
	local tq2QlFrame = CreateFrame("Frame", "tomQuest2QlFrame", UIParent)
	tinsert(UISpecialFrames, "tomQuest2QlFrame")
	tq2QlFrame:SetWidth(640)
	tq2QlFrame:SetHeight(512)
	tq2QlFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
	tq2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])

	local close = CreateFrame("Button", nil, tq2QlFrame, "UIPanelCloseButton")
	close:SetPoint("TOPRIGHT")
	close:SetScript("OnClick", function()
		tq2QlFrame.currentUid = nil
		tq2QlFrame:Hide()
	end)

	local resize = CreateFrame("Button", nil, tq2QlFrame)
	resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
	resize:SetHeight(12)
	resize:SetWidth(12)
	resize:SetPoint("BOTTOMRIGHT", -3, 3)
	self:setFrameTooltip(resize, LI["Click to resize tomQuest2"])

	local titlereg = CreateFrame("Button", nil, tq2QlFrame)
	titlereg:SetPoint("TOPLEFT", 5, -5)
	titlereg:SetPoint("TOPRIGHT", 0, 0)
	titlereg:SetHeight(20)
	titlereg:SetScript("OnMouseDown", function(frame)
		tq2QlFrame:StartMoving()
	end)
	titlereg:SetScript("OnMouseUp", function(frame)
		tq2QlFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2QlFrame")
	end)

	tq2QlFrame:EnableMouse()
	tq2QlFrame:SetMovable(1)
	tq2QlFrame:SetResizable(1)
	tq2QlFrame:SetMinResize(300, 300)
	tq2QlFrame:SetWidth(400)
	tq2QlFrame:SetFrameLevel(0)
	tq2QlFrame:SetWidth(325)
	tq2QlFrame:SetHeight(450)

	tq2QlFrame.close = close
	tq2QlFrame.resize = resize
	tq2QlFrame.titlereg = titlereg

	tq2QlFrame.title = tq2QlFrame:CreateFontString(nil, "ARTWORK")
	tq2QlFrame.title:SetPoint("TOP", 0, -4)

	tq2QlFrame.scroll = CreateFrame("ScrollFrame", "tomQuest2QlScrollFrame", tq2QlFrame, "UIPanelScrollFrameTemplate")
	tq2QlFrame.scroll:SetPoint("TOPLEFT", 30, -30)
	tq2QlFrame.scroll:SetPoint("BOTTOMRIGHT", -35, 55)

	tq2QlFrame.tooltip = CreateFrame("GameTooltip", "tomQuest2QlTooltip", UIParent, "GameTooltipTemplate")

	tq2QlFrame.text = CreateFrame("SimpleHTML", "tomQuest2QlSimpleHTML", tq2QlFrame.scroll)
	tq2QlFrame.text:SetPoint("TOPLEFT", 0, 0)
	tq2QlFrame.text:SetHeight(300)
	tq2QlFrame.text:SetFontObject(GameFontHighlight)

	tq2QlFrame.abandon = CreateFrame("Button", "tq2QlAbandonFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.abandon:SetWidth(125)
	tq2QlFrame.abandon:SetHeight(21)
	tq2QlFrame.abandon:SetPoint("BOTTOMLEFT", tq2QlFrame, "BOTTOMLEFT", 15, 10)
	tq2QlFrame.abandon:SetText(ABANDON_QUEST)

	tq2QlFrame.share = CreateFrame("Button", "tq2QlShareFrame", tq2QlFrame, "UIPanelButtonTemplate")
	tq2QlFrame.share:SetWidth(125)
	tq2QlFrame.share:SetHeight(21)
	tq2QlFrame.share:SetPoint("BOTTOMRIGHT", tq2QlFrame, "BOTTOMRIGHT", -15, 10)
	tq2QlFrame.share:SetText(SHARE_QUEST)
	
	local function tq2QlOnUpdateFix(self, elapsed)
		self.text:SetText(self.updateText)
		self.updateText = nil
		self:SetScript("OnUpdate", nil)
	end

	function tq2QlFrame.text:mySetText(msg)
		self.text = msg
		tq2QlFrame.updateText = msg
		tq2QlFrame:SetScript("OnUpdate", tq2QlOnUpdateFix)
	end

	function tq2QlFrame.text:updateSize()
		local qlw = tomQuest2QlFrame:GetWidth()
		local width = qlw - 65
		self:SetWidth(width)
		self:mySetText(self.text)
	end
	tq2QlFrame.text:SetScript("OnHyperlinkClick", function(...) self:onHyperlinkClick(...) end)
	tq2QlFrame.text:SetScript("OnHyperlinkEnter", function(...) self:onHyperlinkEnter(...) end)
	tq2QlFrame.text:SetScript("OnHyperlinkLeave", function() tq2QlFrame.tooltip:Hide() end)
	tq2QlFrame.scroll:SetScrollChild(tq2QlFrame.text)
	resize:SetScript("OnMouseDown", function(frame)
		tomQuest2QlFrame:StartSizing()
	end)
	resize:SetScript("OnMouseUp", function(frame)
		tomQuest2QlFrame:StopMovingOrSizing()
		self:savePosition("tomQuest2QlFrame")
		tq2QlFrame.text:updateSize()
	end)
	tq2QlFrame:SetAlpha(1)
	tq2QlFrame:Show()
	self:lockUnlockQlFrame()
end

function module:lockUnlockQlFrame()
	local tq2QlFrame = tomQuest2QlFrame
	tq2QlFrame:ClearAllPoints()
	local tq2ParentFrame = tomQuest2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if not tq2ParentFrame then
			tq2ParentFrame = CreateFrame("frame", "tomQuest2ParentFrame", UIParent)
			tq2ParentFrame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
			tq2ParentFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
			tq2ParentFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
			tinsert(UISpecialFrames, "tomQuest2ParentFrame")
			tq2ParentFrame.close = CreateFrame("Button", nil, tq2ParentFrame, "UIPanelCloseButton")
			tq2ParentFrame.close:SetPoint("TOPRIGHT")
			tq2ParentFrame.resize = CreateFrame("Button", nil, tq2ParentFrame)
			tq2ParentFrame.resize:SetNormalTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
			tq2ParentFrame.resize:SetHighlightTexture("Interface\\AddOns\\tomQuest2\\media\\Resize")
			tq2ParentFrame.resize:SetHeight(12)
			tq2ParentFrame.resize:SetWidth(12)
			tq2ParentFrame.resize:SetPoint("BOTTOMRIGHT", -3, 3)
			self:setFrameTooltip(tq2ParentFrame.resize, LI["Click to resize tomQuest2"])
			tq2ParentFrame.resize:SetScript("OnMouseDown", function(frame)
				tq2ParentFrame:StartSizing()
			end)
			tq2ParentFrame.titlereg = CreateFrame("Button", nil, tq2ParentFrame)
			tq2ParentFrame.titlereg:SetPoint("TOPLEFT", 5, -5)
			tq2ParentFrame.titlereg:SetPoint("TOPRIGHT", 0, 0)
			tq2ParentFrame.titlereg:SetHeight(20)
			tq2ParentFrame.titlereg:SetScript("OnMouseDown", function(frame)
				tq2ParentFrame:StartMoving()
			end)
			tq2ParentFrame.titlereg:SetScript("OnMouseUp", function(frame)
				tq2ParentFrame:StopMovingOrSizing()
				self:savePosition("tomQuest2ParentFrame")
			end)
			tq2ParentFrame:EnableMouse()
			tq2ParentFrame:SetMovable(1)
			tq2ParentFrame:SetResizable(1)
			tq2ParentFrame:SetMinResize(600, 300)
			tq2ParentFrame:SetFrameLevel(0)
			tq2ParentFrame:SetWidth(prdb["tomQuest2ParentFrame"].width)
			tq2ParentFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
		end
		tq2ParentFrame:SetAlpha(1)
		tq2ParentFrame:SetFrameStrata("HIGH")
		tq2QlFrame:SetParent(tomQuest2ParentFrame)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		tq2QlFrame.close:Hide()
		tq2QlFrame.resize:Hide()
		tq2QlFrame.titlereg:Hide()
		tq2QlFrame:SetBackdropColor(0, 0, 0, 0)
		tq2QlFrame:SetBackdropBorderColor(0, 0, 0, 0)
		self:restorePosition("tomQuest2ParentFrame")
		tq2QlFrame:SetWidth(prdb["tomQuest2ParentFrame"].width / 2)
		tq2QlFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
		if prdb.attach == 1 then
			tq2QlFrame:SetPoint("TOPLEFT")
			tq2QlFrame:SetPoint("BOTTOMLEFT")
		elseif prdb.attach == 2 then
			tq2QlFrame:SetPoint("TOPRIGHT")
			tq2QlFrame:SetPoint("BOTTOMRIGHT")
		end
		tq2QlFrame.text:updateSize()
	elseif prdb.showQuestLog then
		tq2QlFrame:SetParent(UIParent)
		tq2QlFrame.title:SetFontObject(GameFontHighlight)
		tq2QlFrame.title:SetText("tomQuest2 - Quest Log")
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2QlFrame:SetAlpha(1)
		tq2QlFrame.close:Show()
		tq2QlFrame.resize:Show()
		tq2QlFrame.titlereg:Show()
		tq2QlFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		tq2QlFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		tq2QlFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2QlFrame")
		tq2QlFrame.text:updateSize()
	elseif not prdb.showQuestLog then
		tq2QlFrame:Hide()
	end

end

function module:lockUnlockLhFrame()
	local tq2LhFrame = tomQuest2LhFrame
	local tq2QlFrame = tomQuest2QlFrame
	tq2LhFrame:ClearAllPoints()
	local tq2ParentFrame = tomQuest2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		tq2ParentFrame.close:SetScript("OnClick", function()
			tq2LhFrame.currentUid = nil
			tq2QlFrame.currentUid = nil
			tq2ParentFrame:Hide()
		end)
		tq2ParentFrame.resize:SetScript("OnMouseUp", function(frame)
			tq2ParentFrame:StopMovingOrSizing()
			local width = tq2ParentFrame:GetWidth()/2
			local height = tq2ParentFrame:GetHeight()
			self:savePosition("tomQuest2ParentFrame")
			tq2QlFrame:SetWidth(width)
			tq2LhFrame:SetWidth(width)
			tq2QlFrame:SetHeight(height)
			tq2LhFrame:SetHeight(height)
			tq2QlFrame.text:updateSize()
			tq2LhFrame.text:updateSize()
		end)
		tq2LhFrame:SetParent(tq2ParentFrame)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		tq2LhFrame:SetBackdropColor(0, 0, 0, 0)
		tq2LhFrame:SetBackdropBorderColor(0, 0, 0, 0)
		tq2LhFrame.close:Hide()
		tq2LhFrame.resize:Hide()
		tq2LhFrame.titlereg:Hide()
		tq2LhFrame:SetWidth(prdb["tomQuest2ParentFrame"].width / 2)
		tq2LhFrame:SetHeight(prdb["tomQuest2ParentFrame"].height)
		if prdb.attach == 1 then
			tq2LhFrame:SetPoint("TOPRIGHT")
			tq2LhFrame:SetPoint("BOTTOMRIGHT")
		elseif prdb.attach == 2 then
			tq2LhFrame:SetPoint("TOPLEFT")
			tq2LhFrame:SetPoint("BOTTOMLEFT")
		end
		tq2LhFrame.text:updateSize()
	elseif (prdb.showLightHeaded and LightHeaded) then
		tq2LhFrame:SetParent(UIParent)
		tq2LhFrame.title:SetFontObject(GameFontHighlight)
		tq2LhFrame.title:SetText("tomQuest2 - LightHeaded")
		if tq2ParentFrame then
			tq2ParentFrame:Hide()
		end
		tq2LhFrame:SetAlpha(1)
		tq2LhFrame.close:Show()
		tq2LhFrame.resize:Show()
		tq2LhFrame.titlereg:Show()
		tq2LhFrame:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
		tq2LhFrame:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
		tq2LhFrame:SetFrameStrata("HIGH")
		self:restorePosition("tomQuest2LhFrame")
		tq2LhFrame.text:updateSize()
	elseif not (prdb.showLightHeaded and LightHeaded) then
		tq2LhFrame:Hide()
	end
end


function module:savePosition(name)
	local f = getglobal(name)
	prdb[name].point, prdb[name].x, prdb[name].y = "BOTTOMLEFT", f:GetLeft(), f:GetBottom()
	prdb[name].width = f:GetWidth()
	prdb[name].height = f:GetHeight()
end

function module:restorePosition(name)
	local f = getglobal(name)
	f:ClearAllPoints()
	f:SetPoint(prdb[name].point, prdb[name].x, prdb[name].y)
	f:SetWidth(prdb[name].width)
	f:SetHeight(prdb[name].height)
end

function module:setFrameTooltip(frame, text)
	frame:SetScript("OnEnter", function(frame, ...)
		GameTooltip:SetOwner(frame, "ANCHOR_TOPLEFT")
		GameTooltip:SetText(text)
		GameTooltip:Show()
	end)
	frame:SetScript("OnLeave", function(frame, ...)
		GameTooltip:Hide()
	end)
end

function module:updateQuestsLightHeadedInformations(uid, page)
	local tq2LhFrame = tomQuest2LhFrame
	if not tq2LhFrame:IsVisible() then return end

	local lhText
	if #stack > 0 then
		lhText = format("%s%s", "|cFF0066FF|Hlhref:back|h[Back]|h|r\n",module:getQuestLightHeadedInformations(uid, page))
	else
		lhText = module:getQuestLightHeadedInformations(uid, page)
	end
	tq2LhFrame.text:mySetText(lhText)

	local num = LightHeaded:GetNumQuestComments(uid)
	if not num then return end
	tq2LhFrame.page:SetText(format(LI["Page %s of %s"], page, num + 1))

	tq2LhFrame.page:Show()
	tq2LhFrame.prev:Show()
	tq2LhFrame.next:Show()

	if page == 1 then
		tq2LhFrame.prev:Disable()
	else
		tq2LhFrame.prev:Enable()
	end
	if page == num + 1 then
		tq2LhFrame.next:Disable()
	else
		tq2LhFrame.next:Enable()
	end

	tq2LhFrame.currentUid = uid
	tq2LhFrame.currentPage = page
end


------------------------------
--- LightHeaded & QuestLog ---
------------------------------
function module:displayQuestInformations(event, uid)
	if not core.quests[uid] then return end
	if prdb.showQuestLog and not tomQuest2QlFrame then
		module:createQlGUI()
	end
	if LightHeaded and prdb.showLightHeaded and not tomQuest2LhFrame then
		module:createLhGUI()
	end
	local tq2QlFrame = tomQuest2QlFrame
	local tq2LhFrame = tomQuest2LhFrame
	local tq2ParentFrame = tomQuest2ParentFrame
	if prdb.showQuestLog and prdb.showLightHeaded and LightHeaded and (prdb.attach == 1 or prdb.attach == 2) then
		if tq2ParentFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2ParentFrame:Hide()
			tq2QlFrame.currentUid = nil
			tq2LhFrame.currentUid = nil
			return
		end
		tq2ParentFrame:Show()
	else
		if tq2QlFrame and tq2QlFrame:IsVisible() and tq2QlFrame.currentUid and tq2QlFrame.currentUid == uid then
			tq2QlFrame:Hide()
			tq2QlFrame.currentUid = nil
			return
		elseif tq2LhFrame and tq2LhFrame:IsVisible() and tq2LhFrame.currentUid and tq2LhFrame.currentUid == uid then
			tq2LhFrame:Hide()
			tq2LhFrame.currentUid = nil
			return
		end
	end

	if prdb.showQuestLog then
		tq2QlFrame:Show()
		tq2QlFrame.currentUid = uid
		local fontN = CreateFont("tq2NFont")
		fontN:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSize)
		local fontH1 = CreateFont("tq2H1Font")
		fontH1:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSizeH1, "OUTLINE")
		local fontH2 = CreateFont("tq2H2Font")
		fontH2:SetFont(LSM:Fetch("font", prdb.qlFont), prdb.qlFontSizeH2, "OUTLINE")
		tq2QlFrame.title:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject(fontN)
		tq2QlFrame.text:SetFontObject("h1", fontH1)
		tq2QlFrame.text:SetFontObject("h2", fontH2)

		if core.quests[uid].complete[core.playerName] and core.quests[uid].id then
			tq2QlFrame.abandon:Enable()
			tq2QlFrame.abandon:SetScript("OnClick", function()
				if core.quests[uid].complete[core.playerName] and core.quests[uid].id then
					local wasSelected = GetQuestLogSelection()
					SelectQuestLogEntry(core.quests[uid].id)
					SetAbandonQuest()
					local items = GetAbandonQuestItems()
					if items then
						StaticPopup_Hide("ABANDON_QUEST")
						StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
					else
						StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
						StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
					end
					SelectQuestLogEntry(wasSelected)
				end
			end)
		else
			tq2QlFrame.abandon:Disable()
		end

 		if core.numPartyMembers > 0 and core.quests[uid].complete[core.playerName] and core.quests[uid].id and core.quests[uid].sharable then
			tq2QlFrame.share:Enable()
			tq2QlFrame.share:SetScript("OnClick", function()
				local wasSelected = GetQuestLogSelection()
				SelectQuestLogEntry(core.quests[uid].id)
				QuestLogPushQuest()
				SelectQuestLogEntry(wasSelected)
			end)
		else
			tq2QlFrame.share:Disable()
 		end

		local qlText
		if not core.quests[uid].description then
			qlText = format("<HTML><BODY><H1>|cff%s%s%s|r %s</H1><BR/><P>%s</P></BODY></HTML>", core.quests[uid].color, core.quests[uid].tag, core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("&lt;%d&gt;", core.quests[uid].playersOnQuest) or "", LI["Quests Informations not yet available, please try once synchronization with party is complete"])
		else
			local rewards = ""
			if (core.quests[uid].titleReward ~= "" or core.quests[uid].rewardSpellName ~= "" or (core.quests[uid].numRewards + core.quests[uid].numChoices + core.quests[uid].money + core.quests[uid].honor + core.quests[uid].talent) > 0 ) then
				local money, honor, talent, title, spell, itemReward, itemChoice = "", "", "", "", "", "" ,""
				if core.quests[uid].money > 0 then
					local gold = floor(abs(core.quests[uid].money) / 10000)
					local silver = mod(floor(abs(core.quests[uid].money) / 100), 100)
					local copper = mod(floor(abs(core.quests[uid].money)), 100)
					if gold > 0 then
						if silver > 0 then
							if copper > 0 then
								money = format("%s |cff%s%d|r.|cff%s%d|r.|cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_GOLD, gold, COLOR_SILVER, silver, COLOR_COPPER, copper)
							else
								money = format("%s |cff%s%d|r.|cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_GOLD, gold, COLOR_SILVER, silver)
							end
						elseif copper > 0 then
							money = format("%s |cff%s%d|r.|cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_GOLD, gold, COLOR_COPPER, copper)
						else
							money = format("%s |cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_GOLD, gold)
						end
					elseif silver > 0 then
						if copper > 0 then 
							money = format("%s |cff%s%d|r.|cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_SILVER, silver, COLOR_COPPER, copper)
						else
							money = format("%s |cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_SILVER, silver)
						end
					else
						money = format("%s |cff%s%d|r<BR/>", LI["You will rececive:"], COLOR_COPPER, copper)
					end
				end
				if core.quests[uid].honor > 0 then
					honor = format(LI["Honor points: %s<BR/>"], core.quests[uid].honor)
				end
				if core.quests[uid].talent > 0 then
					talent = format(LI["Talent points: %d<BR/>"], core.quests[uid].talent)
				end
				if core.quests[uid].titleReward ~= "" then
					title = format(LI["Title: %s<BR/>"], core.quests[uid].titleReward)
				end
				if core.quests[uid].rewardSpellName ~= "" then
					spell = format("|T%s:32|t %s", core.quests[uid].rewardSpellTexture, core.quests[uid].rewardSpellName)
				end
				if core.quests[uid].numRewards > 0 then
					for i = 1, core.quests[uid].numRewards do
						local numItems = core.quests[uid].itemRewards[i].numItems == 1 and "" or format(" %d",core.quests[uid].itemRewards[i].numItems)
						if core.quests[uid].itemRewards[i].texture then
							itemReward = format("%s%s|T%s:32|t%s<BR/>", itemReward, numItems, core.quests[uid].itemRewards[i].texture, core.quests[uid].itemRewards[i].link)
						else
							itemReward = format("%s%s%s<BR/>", itemReward, numItems, core.quests[uid].itemRewards[i].link)
						end
					end
				end
				if core.quests[uid].numChoices > 0 then
					for i = 1, core.quests[uid].numChoices do
						local numItems = core.quests[uid].itemChoices[i].numItems == 1 and "" or format(" %d", core.quests[uid].itemChoices[i].numItems)
						if core.quests[uid].itemChoices[i].texture then
							itemChoice = format("%s%s|T%s:32|t%s<BR/>", itemChoice, numItems, core.quests[uid].itemChoices[i].texture, core.quests[uid].itemChoices[i].link)
						else
							itemChoice = format("%s%s%s<BR/>", itemChoice, numItems, core.quests[uid].itemChoices[i].link)
						end
					end
					itemChoice = format(LI["Choose one of the following reward:<BR/>%s"], itemChoice)
				end
				rewards = format("<BR/><H2>%s</H2><P> </P><P>%s%s%s%s%s%s%s</P>", LI["Rewards:"], money, honor, talent, title, spell, itemReward, itemChoice)
			end
			local objectives = ""
			if core.questsObjectives[uid] then
				for objective, qoData in pairs(core.questsObjectives[uid]) do
					objectives = format("%s|cff%s|Hqlref:qo:%d:%s|h%s|h|r %s<BR/>", objectives, qoData.qoColor, uid, qoData.desc, qoData.desc, qoData.qoStatus)
				end
				objectives = format("%s<BR/>", objectives)
			end
 			qlText = format("<HTML><BODY><H1>|cff%s%s%s|r %s</H1><BR/><P>%s</P><BR/><P>%s</P><H2>%s</H2><BR/><P>%s</P>%s</BODY></HTML>", core.quests[uid].color, core.quests[uid].tag, core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("&lt;%d&gt;", core.quests[uid].playersOnQuest) or "", gsub(gsub(core.quests[uid].objective, "<", "&lt;"), ">", "&gt;"), objectives, LI["Description:"], gsub(gsub(core.quests[uid].description, "<", "&lt;"), ">", "&gt;"), rewards)
		end
		tq2QlFrame.text:mySetText(qlText)
	end

	if LightHeaded and prdb.showLightHeaded then
		tq2LhFrame:Show()
		local font = CreateFont("tq2LhFont")
		font:SetFont(LSM:Fetch("font", prdb.lhFont), prdb.lhFontSize)
		tq2LhFrame.title:SetFontObject(font)
		tq2LhFrame.page:SetFontObject(font)
		tq2LhFrame.text:SetFontObject(font)
		tq2LhFrame.currentUid = uid
		tq2LhFrame.currentPage = 1
		local num = LightHeaded:GetNumQuestComments(uid) or 0
		if prdb.showAllCommentsOnOnePage then
			tq2LhFrame.next:Hide()
			tq2LhFrame.prev:Hide()
			tq2LhFrame.page:Hide()
			local lhText = module:getQuestLightHeadedInformations(uid, 0, num)
			if lhText then
				tq2LhFrame.text:mySetText(lhText)
			end
		else
			tq2LhFrame.page:Show()
			local lhText = module:getQuestLightHeadedInformations(uid, 1, num)
			if lhText then
				tq2LhFrame.page:SetText(format(LI["Page %s of %s"], 1, num + 1))
				tq2LhFrame.text:mySetText(lhText)
				if num == 0 then
					tq2LhFrame.prev:Hide()
					tq2LhFrame.next:Hide()
				else
					tq2LhFrame.prev:Show()
					tq2LhFrame.next:Show()
					tq2LhFrame.prev:Disable()
					if tq2LhFrame.currentPage == num + 1 then
						tq2LhFrame.next:Disable()
					else
						tq2LhFrame.next:Enable()
					end
				end
			end
		end
	end
end

function module:getQuestLightHeadedInformations(uid, page, numComments)
	local text = ""
	local title = core.quests[uid] and core.quests[uid].title or LightHeaded:GetQuestName(uid)
	if page == 0 or page == 1 then
		for uid, sharable, level, reqlev, stype, sname, sid, etype, ename, eid, exp, rep, series in LightHeaded:IterateQuestInfo(uid) do
			if reqlev ~= "" then
				reqlev = format(LI["%s|c%sRequired Level:|r %d"], "\n", prdb.color, reqlev)
			end
			if stype ~= "" and sname ~= "" then
				if stype == "npc" then
					sname = format(LI["%s|c%sStarts:|r |cFF0066FF|Hlhref:npc:%s:%s|h[%s]|h|r"], "\n", prdb.color, sid, sname, sname)
				else
					sname = format(LI["%s|c%sStarts:|r |cFFFFFFFF%s|r"], "\n", prdb.color, sname)
				end
			end
			if etype ~= "" and ename ~= "" then
				if etype == "npc" then
					ename = format(LI["%s|c%sEnds:|r |cFF0066FF|Hlhref:npc:%s:%s|h[%s]|h|r"], "\n", prdb.color, eid, ename, ename)
				else
					ename = format(LI["%s|c%sEnds:|r |cFFFFFFFF%s|r"], "\n", prdb.color, ename)
				end
				
			end
			if exp ~= "" then
				exp = format(LI["%s|c%sExperience:|r %d"], "\n", prdb.color, exp)
			end
			local repText = ""
			if rep ~= "" then
				repText = format(LI["%s|c%sReputation Gains:|r"], "\n", prdb.color)
				for name, value in rep:gmatch("([^\029]+)\029([^\029]+)") do
					repText = format("%s\n|c%s - %s |r: %s", repText, prdb.color, name, value)
				end
			end
			local seriesText = ""
			if series ~= "" then
				seriesText = format(LI["%s|c%sQuest Series:|r"], "\n", prdb.color)
				for step, id in series:gmatch("([^\029]+)\029([^\029]+)") do
					if uid == id then
						seriesText = format("%s\n|c%s%s - %s |r", seriesText, prdb.color, step, title)
					else
						seriesText = format("%s\n|c%s%s - |r|cFF0066FF|Hlhref:quest:%s|h[%s]|h|r", seriesText, prdb.color, step, id, LightHeaded:GetQuestName(id))
					end
				end
			end
			text = format(LI["|c%sName:|r %s%s|c%sQuest ID:|r |Hlhref:external:http://www.wowhead.com/?quest=%d|h%d|h%s|c%sSharable:|r %s%s|c%sLevel:|r %s%s%s%s%s%s%s"], prdb.color, title, "\n", prdb.color, uid, uid, "\n", prdb.color, sharable and LI["Yes"] or LI["No"], "\n", prdb.color, level, reqlev, sname, ename, exp, repText, seriesText)
		end
	end
	if page > 1 then
		text = format(LI["|c%sComments for:|r %s%s%s"], prdb.color, title, "\n", LightHeaded:GetCommentText(uid, page - 1))
	elseif page == 0 and numComments > 0 then
		text = format(LI["%s%s|c%sComments for:|r %s"], text, "\n\n", prdb.color, title)
		for page = 1, numComments do
			text = format("%s\n\n%s", text, LightHeaded:GetCommentText(uid, page))
		end
	end
	return text
end

function module:onHyperlinkClick(frame, link)
	local tq2LhFrame = tomQuest2LhFrame
	if not link then return end
	if link:match("^qlref:qo") then
		local uid, objective = link:match("^qlref:qo:(.+):(.+)$")
		if Cartographer_QuestInfo then
			local q = Cartographer_QuestInfo:GetQuest(uid)
			if not q then return end
			if IsControlKeyDown() then
				if q.start_npc then
					data.quest = q.title_full
					data.obj = nil
					data.npc = q.start_npc
					Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
				end
			elseif IsShiftKeyDown() then
				if q.end_npc then
					data.quest = q.title_full
					data.obj = nil
					data.npc = q.end_npc
					Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
				end
			else
				for _, obj in pairs(q.objs) do
					if obj.title:find(objective, 1, true) then
						data.quest = q.title_full
						data.obj = obj
						data.npc = nil
					end
				end
				if data.obj ~= nil then
					Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
				end
			end
		elseif MobMapDotParentFrame and MobMap_ParseQuestObjective then
			MobMap_ParseQuestObjective(objective)
		end
	elseif link:match("^lhref:npc") then
		tinsert(stack, tq2LhFrame.currentPage)
		tinsert(stack, tq2LhFrame.currentUid)
		local id, name = link:match("^lhref:npc:(.+):(.+)$")
		id = tonumber(id)
		local text = format("|cFF0066FF|Hlhref:back|h[Back]|h|r\n\n\n%s|cFFFFFFFF%s|r:\n", LI["Known locations for "], name)
		if LightHeaded:LoadNPCData(id) then
			for c,z,x,y in LightHeaded:IterateNPCLocs(id) do
				text = format("%s\n|cFF0066FF|Hlhref:zcoord:%s:%s:%s:%s:%s|h[%s]|h|r", text, c, z, x, y, name, LightHeaded:GetCoordText(x, y))
			end
		else
			text = format("%s\n%s", text, LI["Location unknown"])
		end
		tq2LhFrame.text:mySetText(text)
		tq2LhFrame.next:Disable()
		tq2LhFrame.prev:Disable()
	elseif link:match("^lhref:coord") then
		local xt, yt, uid = link:match("^lhref:coord:([^:]+):([^:]+):(.+)")
		local x = tonumber(xt)
		local y = tonumber(yt)
		if not x or not y then
			core:Print(format(LI["The coordinate you clicked on was not valid: %s"], link))
			return
		end
		local note
		if not tonumber(uid) then
			note = uid
		else
			note = format("[LH] %s", LightHeaded:GetQuestName(uid))
		end
		if TomTom then
			TomTom:AddWaypoint(x, y, note)
		elseif Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(nil, nil, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Local) == "function" then
			MN_ThottInterface_Local(x, y, note)
		elseif Cartographer_Waypoints then
			core:Print(LI["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:zcoord") then
		local c, z, x, y, uid = link:match("^lhref:zcoord:([^:]+):([^:]+):([^:]+):([^:]+):(.+)")
		c, z, x, y = tonumber(c), tonumber(z), tonumber(x), tonumber(y)

		local note
		if not tonumber(uid) then
			note = uid
		else
			note = format("[LH] %s", LightHeaded:GetQuestName(uid))
		end

		if TomTom and type(TomTom.AddZWaypoint) == "function" then
			TomTom:AddZWaypoint(c, z, x, y, note)
		elseif  Cartographer_Waypoints and type(Cartographer_Waypoints.AddLHWaypoint) == "function" then
			Cartographer_Waypoints:AddLHWaypoint(c, z, x, y, note)
		elseif MN_ThottInterface and type(MN_ThottInterface_Legacy) == "function" then
			MN_ThottInterface_Legacy(c, z, x, y, note)
		elseif TomTom then
			core:Print(LI["NPC locations are not supported in this version of TomTom."])
		elseif Cartographer_Waypoints then
			core:Print(LI["LightHeaded waypoints are not supported in this version of Cartographer_Waypoints."])
		end
	elseif link:match("^lhref:external") then
		local url = link:match("^lhref:external:(.+)$")
		ChatFrameEditBox:Show()
		ChatFrameEditBox:SetText(url)
		ChatFrameEditBox:SetFocus()
	elseif link:match("^lhref:quest:") then
		tinsert(stack, tq2LhFrame.currentPage)
		tinsert(stack, tq2LhFrame.currentUid)
		local uid = link:match("^lhref:quest:(.+)$")
		uid = tonumber(uid)
		module:updateQuestsLightHeadedInformations(uid, 1)
	elseif link:match("^lhref:back$") then
		local uid = tremove(stack)
		local page = tremove(stack)
		module:updateQuestsLightHeadedInformations(uid, page)
	elseif link:match("^lhref:item:") or link:match("^item:") then
		if link:match("^lhref:item:") then
			link = link:sub(7)
		end
		local name, chatLink = GetItemInfo(link)
		if name then
			if IsControlKeyDown() then
				DressUpItemLink(link)
			elseif IsShiftKeyDown() and ChatFrameEditBox:IsVisible() then
				ChatFrameEditBox:Insert(chatLink)
			else
				SetItemRef(link)
			end
		elseif IsAltKeyDown() then
			local tooltip = tq2LhFrame.tooltip
			tooltip.linkForced = link
			tooltip:SetOwner(tq2LhFrame.text, "ANCHOR_CURSOR")
			tooltip:SetHyperlink(link)
		end
	else
		core:Print(format(LI["Unmanaged lhref:%s "], link))
	end
end

function module:onHyperlinkEnter(frame, link)
	local tq2LhFrame = tomQuest2LhFrame
	local tq2QlFrame = tomQuest2QlFrame
	if not link then return end
	if link:match("^lhref:item") then
		link = link:sub(7)
		tq2LhFrame.tooltip:SetOwner(frame,"ANCHOR_CURSOR")
		tq2LhFrame.tooltip:SetHyperlink(link)
		tq2LhFrame.tooltip:Show()
	elseif link:match("^item") then
		tq2QlFrame.tooltip:SetOwner(frame,"ANCHOR_CURSOR")
		tq2QlFrame.tooltip:SetHyperlink(link)
		tq2QlFrame.tooltip:Show()
	end
end