-----------------------
--- Module Creation ---
-----------------------
local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")

-----------------------
--- Library loading ---
-----------------------
local LQTo = LibStub("AceLocale-3.0"):GetLocale("tomQuest2-QuestsTooltip")
local LSM = LibStub("LibSharedMedia-3.0")
local LibQTip = LibStub("LibQTip-1.0")
local scrollLibQTip = nil
if LibQTip.tipPrototype.UpdateScrolling then
	scrollLibQTip = true
end
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})

--------------------------------------------
--- Copy global functions to local scope ---
--------------------------------------------
local pairs, unpack, tonumber = pairs, unpack, tonumber
local format, sub = string.format, string.sub
local ToggleFrame = ToggleFrame
local ShowUIPanel = ShowUIPanel
local QuestLogPushQuest = QuestLogPushQuest
local MouseIsOver = MouseIsOver
local ItemRefTooltip = ItemRefTooltip
local ChatEdit_InsertLink = ChatEdit_InsertLink
local SelectQuestLogEntry, GetQuestLogSelection = SelectQuestLogEntry, GetQuestLogSelection
local GetQuestLink = GetQuestLink
local CreateFrame, CreateFont, SetFont = CreateFrame, CreateFont, SetFont
local IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown = IsShiftKeyDown, IsControlKeyDown, IsAltKeyDown
local GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest = GetAbandonQuestItems, GetAbandonQuestName, SetAbandonQuest

-----------------
--- Variables ---
-----------------
module.modName = LQTo["Broker Quests Tooltip"]
local _
local data = {}
local prdb, chdb, options
local tip, tipAnchor
local tooltipIsShown = false
local maxQuests = MAX_QUESTS
local timer = CreateFrame("Frame")
local hToFontTQ2 = CreateFont("hToFontTQ2")
local qToFontTQ2 = CreateFont("qToFontTQ2")
local qoToFontTQ2 = CreateFont("qoToFontTQ2")
local outlines = {[""] = LQTo["None"], ["OUTLINE"] = LQTo["Outline"], ["THICKOUTLINE"] = LQTo["Thick Outline"]}
local hideObjectives = {[1] = LQTo["When you've completed the objective"], [2] = LQTo["When everyone in the party have completed the objective"], [3] = LQTo["Never"]}
local defaultColors = {"defaultHeaderNameColor", "defaultHeaderLevelColor", "defaultQuestNameColor", "defaultQuestLevelColor", "defaultQuestObjectiveColor"}

-----------------------
--- Default profile ---
-----------------------
local defaults = {
	profile = {
		showNumCompletedQuests = true,
		showNumQuests = true,
		showMaxQuest = true,
		hideCompletedObjective = 2,
		headerFont = LSM.DefaultMedia.font,
		headerFontSize = 12,
		headerFontOutline = "", 
		questFont = LSM.DefaultMedia.font,
		questFontSize = 10,
		questFontOutline = "",
		questObjectiveFont = LSM.DefaultMedia.font,
		questObjectiveFontSize = 8, 
		questObjectiveFontOutline = "",
		maxWidth = 300,
		hasMaxHeight = true,
		maxHeight = 600,
		scale = 1,
		showQuestItemIcon = true,
		questItemIconSize = 24,
		questsCellPadding = 5,
		objectivesCellPadding = 10,
		collapseButtonSize = 24,
		colorizeHeaderName = true,
		defaultHeaderNameColor = "ffffff",
		hideZoneLevelRange = false,
		colorizeHeaderLevel = true,
		defaultHeaderLevelColor = "ffffff",
		colorizeQuestName = true,
		defaultQuestNameColor = "ffffff",
		colorizeQuestLevel = true,
		defaultQuestLevelColor = "ffffff",
		colorizeQuestObjective = true,
		defaultQuestObjectiveColor = "ffffff",
		colorizeQuestObjectiveCounter = true,
		showHints = true,
		chatLink  = true,
		questsObjectives = true,
		questsMacro = true,
		clickTooltip = true,
		shareQuest = true,
		trackQuest = true,
		abandonQuest = true,
		questLog = true,
		overTooltip = true,
		cqiObjectiveStartNPC = true,
		cqiObjectiveEndNPC = true,
		cqiObjective = true,
		mmObjective = true,
	}
}

---------------
--- Options ---
---------------
local optGetter, optSetter, colorGetter, colorSetter, hexColorGetter, hexColorSetter, ldbTextSetter, headFontSetter, questFontSetter, qoFontSetter, hideItemIcon, hideHNColor, hideZLColor, hideZLColor2, hideQNColor, hideQLColor, hideQOColor, hideCQI, hideMM, hideMaxHeight, hideMaxHeight2
do
	function optGetter(info)
		return prdb[info[#info]]
	end
	function optSetter(info, value)
		prdb[info[#info]] = value
	end
	function colorGetter(info)
		return unpack(prdb[info[#info]])
	end
	function colorSetter(info, arg1, arg2, arg3, arg4)
		prdb[info[#info]][1] = arg1
		prdb[info[#info]][2] = arg2
		prdb[info[#info]][3] = arg3
		prdb[info[#info]][4] = arg4
	end
	function hexColorGetter(info)
		return tonumber(sub(prdb[info[#info]], 1, 2), 16) / 255, tonumber(sub(prdb[info[#info]], 3, 4), 16) / 255, tonumber(sub(prdb[info[#info]], 5, 6), 16) / 255
	end
	function hexColorSetter(info, arg1, arg2, arg3)
		prdb[info[#info]] = format("%02x%02x%02x", arg1*255, arg2*255, arg3*255)
	end
	function ldbTextSetter(info, value)
		prdb[info[#info]] = value
		module:questLogUpdated(nil, true, false)
	end
	function headFontSetter(info, value)
		prdb[info[#info]] = value
		hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	end
	function questFontSetter(info, value)
		prdb[info[#info]] = value
		qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	end
	function qoFontSetter(info, value)
		prdb[info[#info]] = value
		qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
	end
	function hideItemIcon()
		if prdb.showQuestItemIcon then
			return false
		else
			return true
		end
	end
	function hideHNColor()
		if not prdb.colorizeHeaderName then
			return false
		else
			return true
		end
	end
	function hideZLColor()
		if not prdb.hideZoneLevelRange then
			return false
		else
			return true
		end
	end
	function hideZLColor2()
		if (prdb.hideZoneLevelRange and not prdb.colorizeHeaderLevel) or (prdb.colorizeHeaderLevel) then
			return true
		else
			return false
		end
	end
	function hideQNColor()
		if not prdb.colorizeQuestName then
			return false
		else
			return true
		end
	end
	function hideQLColor()
		if not prdb.colorizeQuestLevel then
			return false
		else
			return true
		end
	end
	function hideQOColor()
		if not prdb.colorizeQuestObjective then
			return false
		else
			return true
		end
	end
	function hideCQI()
		if Cartographer_QuestInfo then
			return false
		else
			return true
		end
	end
	function hideMM()
		if MobMapDotParentFrame then
			return false
		else
			return true
		end
	end
	function hideMaxHeight()
		if scrollLibQTip then
			return false
		else
			return true
		end
	end
	function hideMaxHeight2()
		if scrollLibQTip or not prdb.hasMaxHeight then
			return false
		else
			return true
		end
	end
end

local function getOptions()
	if not options then
		options = {
			type = "group",
			name = module.modName,
			desc = module.modName,
			get = optGetter,
			set = optSetter,
			args = {
				showNumCompletedQuests = {
					name = LQTo["Show number of completed quest"],
					type = "toggle",
					set = ldbTextSetter,
					width = "full",
					order = 0,
				},
				showNumQuests = {
					name = LQTo["Show number of quests"],
					type = "toggle",
					set = ldbTextSetter,
					width = "full",
					order = 1,
				},
				showMaxQuest = {
					name = LQTo["Show maximum quests number"],
					type = "toggle",
					set = ldbTextSetter,
					width = "full",
					order = 2,
				},
				tooltipHeader = {
					type = "header",
					name = LQTo["Tooltip Settings"],
					order = 3,
				},
				hideCompletedObjective = {
					name = LQTo["Hide completed quests Objectives"],
					type = "select",
					values = hideObjectives,
					width = "full",
					order = 4,
				},
				headerFont = {
					name = LQTo["Zone header font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = headFontSetter,
					width = "full",
					order = 5,
				},
				headerFontSize = {
					name = LQTo["Zone header font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = headFontSetter,
					width = "full",
					order = 6,
				},
				headerFontOutline = {
					name = LQTo["Zone header font outline"],
					type = "select",
					values = outlines,
					set = headFontSetter,
					width = "full",
					order = 7,
				},
				questFont = {
					name = LQTo["Quests font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = questFontSetter,
					width = "full",
					order = 8,
				},
				questFontSize = {
					name = LQTo["Quests font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = questFontSetter,
					width = "full",
					order = 9,
				},
				questFontOutline = {
					name = LQTo["Quests font outline"],
					type = "select",
					values = outlines,
					set = questFontSetter,
					width = "full",
					order = 10,
				},
				questObjectiveFont = {
					name = LQTo["Quests objectives font"],
					type = "select",
					dialogControl = 'LSM30_Font',
					values = AceGUIWidgetLSMlists.font,
					set = qoFontSetter,
					width = "full",
					order = 11,
				},
				questObjectiveFontSize = { 
					name = LQTo["Quests objectives font size"],
					type = "range",
					min = 6,
					max = 20,
					step = 1,
					set = qoFontSetter,
					width = "full",
					order = 12,
				},
				questObjectiveFontOutline = {
					name = LQTo["Quests objectives font outline"],
					type = "select",
					values = outlines,
					set = qoFontSetter,
					width = "full",
					order = 13,
				},
				maxWidth = {
					name = LQTo["Quests Tooltip column max width"],
					type="range",
					min = 50,
					max = 600,
					step = 5,
					width = "full",
					order = 14,
				},
				hasMaxHeight = {
					name = LQTo["Set quests tooltip maximum height after which it will be scrollable"],
					hidden = hideMaxHeight,
					type = "toggle",
					width = "full",
					order = 15,
				},
				maxHeight = {
					name = LQTo["Quests tooltip max height"],
					hidden = hideMaxHeight2,
					type = "range",
					min = 100,
					max = 2000,
					step = 5,
					width = "full",
					order = 16,
				},
				scale = {
					name = LQTo["Quests Tooltip scale"],
					type="range",
					min = 0,
					max = 1,
					step = 0.1,
					width = "full",
					order = 17,
				},
				showQuestItemIcon = {
					name = LQTo["Show special quest item icon behind quest name"],
					type = "toggle",
					width = "full",
					order = 18,
				},
				questItemIconSize = {
					name = LQTo["Quest item icon size"],
					type = "range",
					hidden = hideItemIcon,
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 19,
				},
				questsCellPadding = {
					name = LQTo["Quests cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 20,
				},
				objectivesCellPadding = {
					name = LQTo["Quests objectives cell padding"],
					type="range",
					min = 0,
					max = 50,
					step = 1,
					width = "full",
					order = 21,
				},
				collapseButtonSize = {
					name = LQTo["Collapse Button Size"],
					type = "range",
					min = 10,
					max = 40,
					step = 1,
					width = "full",
					order = 22,
				},
				colorizeHeaderName = {
					name = LQTo["Colorize Quests Header Name"],
					type = "toggle",
					width = "full",
					order = 23,
				},
				defaultHeaderNameColor = {
					name = LQTo["Default header name color"],
					type = "color",
					hasAlpha = false,
					hidden = hideHNColor,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 24,
				},
				hideZoneLevelRange = {
					name = LQTo["Hide zone level range"],
					type = "toggle",
					width = "full",
					order = 25,
				},
				colorizeHeaderLevel = {
					name = LQTo["Colorize Quests Header Level"],
					hidden = hideZLColor,
					type = "toggle",
					width = "full",
					order = 26,
				},
				defaultHeaderLevelColor = {
					name = LQTo["Default header level color"],
					type = "color",
					hidden = hideZLColor2,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 27,
				},
				colorizeQuestName = {
					name = LQTo["Colorize Quests Name"],
					type = "toggle",
					width = "full",
					order = 28,
				},
				defaultQuestNameColor = {
					name = LQTo["Default quest name color"],
					type = "color",
					hidden = hideQNColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 29,
				},
				colorizeQuestLevel = {
					name = LQTo["Colorize Quests Level"],
					type = "toggle",
					width = "full",
					order = 30,
				},
				defaultQuestLevelColor = {
					name = LQTo["Default quest level color"],
					type = "color",
					hidden = hideQLColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 31,
				},
				colorizeQuestObjective = {
					name = LQTo["Colorize Quests Objectives"],
					type = "toggle",
					width = "full",
					order = 32,
				},
				defaultQuestObjectiveColor = {
					name = LQTo["Default quest objective color"],
					type = "color",
					hidden = hideQOColor,
					hasAlpha = false,
					get = hexColorGetter,
					set = hexColorSetter,
					width = "full",
					order = 33,
				},
				colorizeQuestObjectiveCounter = {
					name = LQTo["Colorize quests objectives counter"],
					type = "toggle",
					width = "full",
					order = 34,
				},
				tooltipInteraction = {
					type = "header",
					name = LQTo["Tooltip interaction"],
					order = 35,
				},
				showHints = {
					name = LQTo["Show interractivity hints at the bottom of the tooltip"],
					type = "toggle",
					width = "full",
					order = 36,
				},
				chatLink  = {
					name = LQTo["Insert quest link in chat on shift right click"],
					type = "toggle",
					width = "full",
					order = 37,
				},
				questsObjectives = {
					name = LQTo["Insert quest objectives in chat on ctrl right click"],
					type = "toggle",
					width = "full",
					order = 38,
				},
				questsMacro = {
					name = LQTo["Update special quest item macro"],
					type = "toggle",
					width = "full",
					order = 39,
				},
				clickTooltip = {
					name = LQTo["Display of quest tooltip on right click"],
					type = "toggle",
					width = "full",
					order = 40,
				},
				shareQuest = {
					name = LQTo["Share quest on ctrl left click"],
					type = "toggle",
					width = "full",
					order = 41,
				},
				trackQuest = {
					name = LQTo["Track quest on shift left click"],
					type = "toggle",
					width = "full",
					order = 42,
				},
				abandonQuest = {
					name = LQTo["Abandon quest on alt left click"],
					type = "toggle",
					width = "full",
					order = 43,
				},
				questLog = {
					name = LQTo["Diplay quest log on left click"],
					type = "toggle",
					width = "full",
					order = 44,
				},
				overTooltip = {
					name = LQTo["Display quest tooltip on quest mouse over"],
					type = "toggle",
					width = "full",
					order = 45,
				},
				cqiObjectiveStartNPC = {
					name = LQTo["Display of start npc in Cartographer quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 46,
				},
				cqiObjectiveEndNPC = {
					name = LQTo["Display of end npc in Cartographer Quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 47,
				},
				cqiObjective = {
					name = LQTo["Display of objective in Cartographer Quest Info"],
					hidden = hideCQI,
					type = "toggle",
					width = "full",
					order = 48,
				},
				mmObjective = {
					name = LQTo["Display of objective in your map with MobMap"],
					hidden = hideMM,
					type = "toggle",
					width = "full",
					order = 49,
				},
			},
		}
	end
	return options
end

----------------------
--- Initialization ---
----------------------
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	self:SetEnabledState(core:getModuleEnabled("questsTooltip"))
	core:registerModuleOptions("questsTooltip", getOptions, module.modName)
	chdb.collapseAll = chdb.collapseAll or 0
	chdb.zoneDisplay = chdb.zoneDisplay or {}
	core.RegisterCallback(self, "updateDBtoV3")
	LSM.RegisterCallback(self, "LibSharedMedia_Registered", "updateUsedMedia")
end

function module:updateDBtoV3(event)
	for k in pairs(defaultColors) do
		if type(prdb[k]) == "table" then
			prdb[k] = format("%02x%02x%02x", (prdb[k][1] or 0)*255, (prdb[k][2] or 0)*255, (prdb[k][3] or 0)*255)
		end
	end
end

local function hideTooltip()
	if not tip or MouseIsOver(tip) then return end
	tooltipIsShown = false
	tip:SetScript("OnUpdate", nil)
	LibQTip:Release(tip)
	tip = nil
end

function module:OnEnable()
	core.RegisterCallback(self, "questLogUpdated")
	hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
	self:Refresh()
end

function module:Refresh()
	prdb = self.db.profile
	chdb = self.db.char
	if not self:IsEnabled() then return end
	hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
	qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
	qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
end

function module:updateUsedMedia(event, mediatype, key)
	if mediatype == "font" then
		local updateTooltip
		if key == prdb.headerFont then
			updateTooltip = true
			hToFontTQ2:SetFont(LSM:Fetch("font", prdb.headerFont), prdb.headerFontSize, prdb.headerFontOutline)
		end
		if key == prdb.questFont then
			updateTooltip = true
			qToFontTQ2:SetFont(LSM:Fetch("font", prdb.questFont), prdb.questFontSize, prdb.questFontOutline)
		end
		if key == prdb.questObjectiveFont then
			updateTooltip = true
			qoToFontTQ2:SetFont(LSM:Fetch("font", prdb.questObjectiveFont), prdb.questObjectiveFontSize, prdb.questObjectiveFontOutline)
		end
		if tooltipIsShown and updateTooltip then
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		end
	end
end

---------------------------
--- Main Module Program ---
---------------------------
function module:questLogUpdated(event, ldbText, tooltip)
	if ldbText then
		core.numQuests = core.numQuests or 0
		core.numCompletedQuests = core.numCompletedQuests or 0
		if prdb.showNumCompletedQuests then
			if prdb.showNumQuests then
				if prdb.showMaxQuest then
					if core.numQuests > 0 then
						dataobj.text = format("|cff%s%s/%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests, maxQuests)
					else
						dataobj.text = format("0/0/%s", maxQuests)
					end
				else
					if core.numQuests > 0 then
						dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests)
					else
						dataobj.text = "0/0"
					end
				end
			elseif prdb.showMaxQuest then
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/maxQuests), core.numCompletedQuests, maxQuests)
			else
				dataobj.text = core.numCompletedQuests
			end
		elseif prdb.showNumQuests then
			if prdb.showMaxQuest then
				dataobj.text = format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numQuests/maxQuests), core.numQuests, maxQuests)
			else
				dataobj.text = core.numQuests
			end	
		elseif prdb.showMaxQuest then
			dataobj.text = maxQuests
		else
			dataobj.text = ""
		end
	end
	if tooltip and tooltipIsShown then
		tip:SetScript("OnLeave", nil)
		LibQTip:Release(tip)
		tip = nil
		module:updateTooltip()
	end
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bType, bValue, bValue2, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if bType == "header" then
			if chdb.collapseAll == 1 then
				chdb.collapseAll = 0
				for zone in pairs(chdb.zoneDisplay) do
					chdb.zoneDisplay[zone] = 0
				end
			else
				chdb.collapseAll = 1
				for zone in pairs(chdb.zoneDisplay) do
					chdb.zoneDisplay[zone] = 1
				end
			end
			tip:SetScript("OnLeave", nil)
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		elseif bType == "zone" then
			if chdb.zoneDisplay[bValue] == 1 then
				chdb.zoneDisplay[bValue] = 0
				chdb.collapseAll = 0
			else
				chdb.zoneDisplay[bValue] = 1
				chdb.collapseAll = 1
			end
			tip:SetScript("OnLeave", nil)
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		elseif bType == "quest" then
			if button == "RightButton" then
				if IsShiftKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.chatLink then
						ChatFrameEditBox:Show()
						ChatEdit_InsertLink(GetQuestLink(core.quests[bValue].id))
					end
				elseif IsControlKeyDown() then
					if core.quests[bValue].numObjectives > 0 and core.quests[bValue].complete[core.playerName] and prdb.questsObjectives then
						ChatFrameEditBox:Show()
						for objective, qoData in pairs(core.questsObjectives[bValue]) do
							local got, _ = qoData.got[core.playerName]:match("^(.-)~(%d-)$")
							ChatFrameEditBox:Insert(format("{%s %s/%s} ", qoData.desc, got, qoData.need))
						end
					end
				elseif IsAltKeyDown() and bType == "quest" and prdb.questsMacro then
					core:generateSpecialItemMacro(bValue)
				elseif core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.clickTooltip then
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetQuestLink(core.quests[bValue].id)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsControlKeyDown() then
					if core.numPartyMembers > 0 and core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and core.quests[bValue].sharable and prdb.shareQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						QuestLogPushQuest()
						SelectQuestLogEntry(wasSelected)
					end
				elseif IsShiftKeyDown() then
					if prdb.trackQuest then
						core.cbh:Fire("toggleTrackedQuest", bValue)
						tip:SetScript("OnLeave", nil)
						LibQTip:Release(tip)
						tip = nil
						module:updateTooltip()
					end
				elseif IsAltKeyDown() then
					if core.quests[bValue].complete[core.playerName] and core.quests[bValue].id and prdb.abandonQuest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bValue].id)
						SetAbandonQuest()
						local items = GetAbandonQuestItems()
						if items then
							StaticPopup_Hide("ABANDON_QUEST")
							StaticPopup_Show("ABANDON_QUEST_WITH_ITEMS", GetAbandonQuestName(), items)
						else
							StaticPopup_Hide("ABANDON_QUEST_WITH_ITEMS")
							StaticPopup_Show("ABANDON_QUEST", GetAbandonQuestName())
						end
						SelectQuestLogEntry(wasSelected)
					end
				else
					core:displayQuestInformations(bValue)
				end
			elseif button == "MiddleButton" then
				if bValue2 ~= nil then
					core.cbh:Fire("toggleQuestInfo", bValue, bValue2)
					if Cartographer_QuestInfo then
						local q = Cartographer_QuestInfo:GetQuest(bValue)
						if not q then return end
						if IsControlKeyDown() and prdb.cqiObjectiveStartNPC then
							if q.start_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.start_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "start", "LeftButton", data)
							end
						elseif IsShiftKeyDown() and prdb.cqiObjectiveEndNPC then
							if q.end_npc then
								data.quest = q.title_full
								data.obj = nil
								data.npc = q.end_npc
								Cartographer_QuestInfo:OnButtonClick(nil, "end", "LeftButton", data)
							end
						elseif prdb.cqiObjective then
							if not q.objs then return end
							for _, obj in pairs(q.objs) do
								if obj.title:find(core.questsObjectives[bValue][bValue2].desc, 1, true) then
									data.quest = q.title_full
									data.obj = obj
									data.npc = nil
								end
							end
							Cartographer_QuestInfo:OnButtonClick(nil, "obj", "LeftButton", data)
						end
					elseif MobMapDotParentFrame and MobMap_ParseQuestObjective and prdb.mmObjective then
						MobMap_ParseQuestObjective(core.questsObjectives[bValue][bValue2].desc)
					end
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overTooltip and bType == "quest" then
			if bValue2 ~= nil and core.questsObjectives[bValue] and core.questsObjectives[bValue][bValue2].qType == "object" and core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop then
				GameTooltip:SetOwner(tip, anchor)
				GameTooltip:AddLine(LQTo["|cffffffffDrop from:|r"])
				for mobName in pairs(core.questsItems[core.questsObjectives[bValue][bValue2].desc].mobDrop) do
					local zones = ""
					for zone in pairs(core.questsMobsItems[mobName].mobLoc) do
						zones = zones == "" and zone or format("%s, %s", zones, zone)
					end
					if core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate then
						GameTooltip:AddDoubleLine(format(LQTo["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones), format("|cffffffff%s%%|r", core.questsMobsItems[mobName].items[core.questsObjectives[bValue][bValue2].desc].dropRate))
					else
						GameTooltip:AddLine(format(LQTo["  |cffffffff%s in|r |cff00ff00%s|r"], mobName, zones))
					end
				end
				GameTooltip:Show()
			elseif core.quests[bValue] and core.quests[bValue].id then
				local link = GetQuestLink(core.quests[bValue].id)
				if link then
					GameTooltip:SetOwner(tip, anchor)
					GameTooltip:SetHyperlink(link)
				end
			end
		end
	end)
	self:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateTooltip()
	tip = LibQTip:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	tooltipIsShown = true
	tip:Clear()
	tip:EnableMouse(true)
	tip:SetScale(prdb.scale)
	local tempLine
	tempLine = tip:AddLine()
	tip:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cffffffff%s|r", chdb.collapseAll == 0 and "Plus" or "Minus", prdb.collapseButtonSize,"tomQuest2"), hToFontTQ2, "CENTER", 2, cProvider, "header", nil, nil, prdb.maxWidth)
	tip:AddLine(" ")
	for zone, zData in pairs(core.questsByZone) do
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, format("|TInterface\\Buttons\\UI-%sButton-Up:%d|t |cff%s%s|r |cff%s%s|r", chdb.zoneDisplay[zone] == 0 and "Plus" or "Minus", prdb.collapseButtonSize, prdb.colorizeHeaderName and zData.color or prdb.defaultHeaderNameColor, zone, prdb.colorizeHeaderLevel and zData.color or prdb.defaultHeaderLevelColor, prdb.hideZoneLevelRange and "" or zData.level ), hToFontTQ2, "LEFT", 2, cProvider, "zone", zone, nil, prdb.maxWidth, 0)
		if chdb.zoneDisplay[zone] == 1 then
			for uid in pairs(zData.quests) do
				tempLine = tip:AddLine()
				if prdb.showQuestItemIcon and core.quests[uid].specialItemIcon then
					if core.quests[uid].tracked == 1 then
						tip:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r |T%s:%d|t|TInterface\\Buttons\\UI-CheckBox-Check:%d|t %s ", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].specialItemIcon, prdb.questItemIconSize, prdb.questItemIconSize, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qToFontTQ2, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, tipAnchor)
					else
						tip:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r |T%s:%d|t %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].specialItemIcon, prdb.questItemIconSize, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qToFontTQ2, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, tipAnchor)
					end
				else
					if core.quests[uid].tracked == 1 then
						tip:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r|TInterface\\Buttons\\UI-CheckBox-Check:%d|t %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, prdb.questItemIconSize, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qToFontTQ2, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, tipAnchor)
					else
						tip:SetCell(tempLine, 1, format("|cff%s%s|r |cff%s%s|r %s", prdb.colorizeQuestLevel and core.quests[uid].color or prdb.defaultQuestLevelColor, core.quests[uid].tag, prdb.colorizeQuestName and core.quests[uid].color or prdb.defaultQuestNameColor, core.quests[uid].title, core.quests[uid].playersOnQuest > 1 and format("<%d>", core.quests[uid].playersOnQuest) or ""), qToFontTQ2, "LEFT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, prdb.questsCellPadding, tipAnchor)
					end
				end
				tip:SetCell(tempLine, 2, core.quests[uid].pqStatus, qToFontTQ2, "RIGHT", 1, cProvider, "quest", uid, nil, prdb.maxWidth, 0, tipAnchor)
				if core.questsObjectives[uid] then
					for objective, qoData in pairs(core.questsObjectives[uid]) do
						if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and qoData.partyComplete == 0) or (prdb.hideCompletedObjective == 1 and qoData.playerComplete == 0) then
							tempLine = tip:AddLine()
							tip:SetCell(tempLine, 1, format("|cff%s%s|r", prdb.colorizeQuestObjective and qoData.qoColor or prdb.defaultQuestObjectiveColor, qoData.desc), qoToFontTQ2, "LEFT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, prdb.objectivesCellPadding, tipAnchor)
							tip:SetCell(tempLine, 2, prdb.colorizeQuestObjectiveCounter and qoData.qoStatus or qoData.qoStatusNC, qoToFontTQ2, "RIGHT", 1, cProvider, "quest", uid, objective, prdb.maxWidth, 0, tipAnchor)
						end
					end
				else
					tempLine = tip:AddLine()
					tip:SetCell(tempLine, 1, core.quests[uid].objective, qoToFontTQ2, "LEFT", 2, cProvider,"quest", uid, nil, prdb.maxWidth, prdb.objectivesCellPadding, tipAnchor)
				end
			end
		end
	end
	tip:AddLine(" ")
	if prdb.showHints then
		if prdb.questLog then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fLeft click|r |cff19ff19Open quests log.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.trackQuest then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fShift + Left click|r |cff19ff19Add quest to tracker.|r"] , qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.shareQuest then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fCtrl + Left click|r |cff19ff19Share quest.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.abandonQuest then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fAlt+ Left click|r |cff19ff19Abandon quest.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.clickTooltip then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fRight click|r |cff19ff19Show quest tooltip.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.chatLink then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fShift + Right click|r |cff19ff19Send quest link to chat.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.questsObjectives then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fCtrl + Right click|r |cff19ff19Send quest progress to chat.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if prdb.questsMacro then
			tempLine = tip:AddLine()
			tip:SetCell(tempLine, 1, LQTo["|cffeda55fAlt + Right click|r |cff19ff19Update special quest item macro.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
		end
		if Cartographer_QuestInfo then
			if prdb.cqiObjectiveStartNPC then
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, LQTo["|cffeda55fCtrl + Middle click|r |cff19ff19On objective to open start npc in Cartographer.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
			end
			if prdb.cqiObjectiveEndNPC then
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, LQTo["|cffeda55fShift + Middle click|r |cff19ff19On objective to open end npc in Cartographer.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
			end
			if prdb.cqiObjective then
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, LQTo["|cffeda55fMiddle click|r |cff19ff19On objective to open it in Cartographer.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
			end
		elseif MobMapDotParentFrame and MobMap_ParseQuestObjective then
			if prdb.mmObjective then
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, LQTo["|cffeda55fMiddle click|r |cff19ff19On objective to open it with MobMap.|r"], qoToFontTQ2, "LEFT", 2, cProvider, "other", nil, nil, prdb.maxWidth)
			end
		end
	end
	if scrollLibQTip and prdb.hasMaxHeight then
		tip:UpdateScrolling(prdb.maxHeight)
	end
	tip:Show()
end

function dataobj.OnEnter(self)
	tipAnchor = self:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	module:updateTooltip()
	tip:SmartAnchorTo(self)
end

function dataobj.OnLeave(self)
	if tip and MouseIsOver(tip) then 
		tip:SetScript("OnUpdate", hideTooltip)
	else
		hideTooltip()
	end
end

function dataobj.OnClick(self,button)
	if button == "RightButton" then
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.Profiles)
		InterfaceOptionsFrame_OpenToCategory(core.optionsFrames.tomQuest2)
	elseif IsShiftKeyDown() then
		tQ2toggleQuestTracker()
	elseif IsAltKeyDown() then
		tQ2toggleAchievementTracker()
	elseif IsControlKeyDown() then
		core:toggleAnnounce()
	else
		ToggleFrame(QuestLogFrame)
	end
end